-- Copyright (C) 2009  Carlo A. Furia
--                     caf@inf.ethz.ch  http://se.inf.ethz.ch/people/furia/
--                     
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.


class
	POSTCONDITION_WEAKENER

inherit
	PRODUCTION_CLASS_NAVIGATOR
	redefine make end

create
	make




feature -- Creation


	make (parse_tree: PRODUCTION_CLASS) is
			-- set parse_tree as the object to navigate
		do
			Precursor (parse_tree)
			create all_candidate_invariants.make
			base_substitution_on_reference
			no_previous_values
			choose_procedure_name (Void)
		end



feature  -- Generation options


	variables_at_a_time: INTEGER
		-- how many variables at a time are substituted?
		-- (default behavior is obtained when set to 0)



	name_based_substitution: BOOLEAN
		-- are variables substituted according to their name?  i.e.,
		-- True iff two occurrences of the same variable are either both
			-- replaced or none of them is



	previous_values: BOOLEAN
			-- consider previous values of variables for substitution?
			-- these values are generated by previous



	base_substitution_on_name is
			-- set name_based_substitution to True
		do
			name_based_substitution := True
		end



	base_substitution_on_reference is
			-- set name_based_substitution to False
		do
			name_based_substitution := False
		end



	try_previous_values is
			-- set previous_values to True
		do
			previous_values := True
		end



	no_previous_values is
			-- set previous_values to False
		do
			previous_values := False
		end



	do_one_variable_at_a_time is
			-- set one_variable_at_a_time to 0
		do
			variables_at_a_time := 0
		end



	do_multiple_variables_at_a_time (how_many: INTEGER) is
			-- set one_variable_at_a_time to how_many
		require
			how_many > 0
		do
			variables_at_a_time := how_many
		end



	choose_procedure_name (a_name: STRING) is
			-- choose `a_name' as the name of the procedure to analyze
		   -- if a_name is Void or empty, the first procedure in the
			-- current file is considered
		do
			if a_name = Void then
				create procedure_name.make_empty
			else
				create procedure_name.make_from_string (a_name)
			end
		end




feature -- Invariant generation


	comment_line: STRING is " // AUTOMATICALLY GENERATED INVARIANT"
			-- string added as comment to every added invariant


	procedure_name: STRING
			-- name of the procedure under analysys

	procedure: PRODUCTION_CLASS_NAVIGATOR
			-- reference to procedure under analysis



	all_candidate_invariants: LINKED_LIST [PRODUCTION_CLASS]
			-- a list of candidate invariants



	set_current_procedure is
			-- set the `procedure' to work on as the first
			-- with name `procedure_name', or to the first in the parse tree
			-- if `procedure_name' is an empty string
			-- if such a procedure does not exist `procedure' is set to Void
		require
			procedure_name /= Void
		do
			from start (ProcedureDecl)
			until
				after
					or item.children[2].children[1].value.is_equal (procedure_name)
					or procedure_name.is_empty
			loop
				forth
			end
			if not after then
				create procedure.make_with_modifies (item, modifies_clauses)
				if procedure_name.is_empty then
					procedure_name.copy (item.children[2].children[1].value)
				end
			else
				procedure := Void
			end
		end



	generate_candidate_invariants is
			-- generate all candidate invariants
			-- for the current procedure in the parsed file
		    -- if no procedure exists in the file
		    -- all_candidate_invariants is set to Void
		local
			found_postcondition: BOOLEAN
		do
			set_current_procedure
			if procedure /= Void then
				-- get handlers to all outer while loops
				while_loops := procedure.outer_while_loops_handler
				from procedure.start (Spec)
				until procedure.after
				loop
					-- look for declarations of postconditions
					if procedure.item.number_of_children = 4
						and procedure.item.children[2].type = ENSURESKW then
						create postcondition.make (procedure.item.children[3].twin)
						found_postcondition := True
					elseif procedure.item.number_of_children = 3
						    and procedure.item.children[1].type = ENSURESKW then
							 create postcondition.make (procedure.item.children[2].twin)
							 found_postcondition := True
					end
					if found_postcondition then
						generate_all_invariants (False)
						if previous_values and variables_at_a_time = 0 then
							generate_all_invariants (True)
						end
						found_postcondition := False
					end
					procedure.forth
				end
			else
				all_candidate_invariants := Void
			end
		end



	add_invariant_to_all_while_loops (an_invariant: PRODUCTION_CLASS) is
			-- add invariant (assumed to be of type Expr)
			-- to every single while loop in the first procedure
		require
			an_invariant.type = Expr
		local
			loop_invariant: PRODUCTION_CLASS
			loop_invariant_star: PRODUCTION_CLASS
		do
			-- build LoopInv block with invariant
			create loop_invariant.make (LoopInv)
			loop_invariant.add_child (create {PRODUCTION_CLASS}.make_with_value
											  (INVARIANTKW, "invariant"))
			loop_invariant.add_child (an_invariant)
			loop_invariant.add_child (create {PRODUCTION_CLASS}.make_with_value
											  (SEMICOLONKW, ";" + comment_line))

			-- for every while loop in procedure
			from procedure.start (Stmt)
			until procedure.after
			loop
				if procedure.item.children[1].type = WHILEKW then
					-- build new LoopInvstar block
					create loop_invariant_star.make (LoopInvstar)
					-- add old LoopInvStar as first child
					loop_invariant_star.add_child (procedure.item.children[5])
					-- add loop_invariant as second child
					loop_invariant_star.add_child (loop_invariant)
					-- replace old LoopInvStar with new one
					procedure.item.children[5] := loop_invariant_star
				end
				procedure.forth
			end
		end



	remove_invariant_from_all_while_loops is
			-- undo the effect of an application of add_invariant_to_all_while_loops
		local
		do
			from procedure.start (Stmt)
			until procedure.after
			loop
				if procedure.item.children[1].type = WHILEKW and
					-- the loop being removed should have been added
					-- by add_invariant_to_all_while_loops
					procedure.item.children[5].number_of_children = 2 and
						procedure.item.children[5].children[2].number_of_children = 3 and
						procedure.item.children[5].children[2].children[3].value.ends_with(comment_line) then
						-- re-attach old LoopInvstar block thus discarding the
						-- new one
						procedure.item.children[5] := procedure.item.children[5].children[1]
				end
				procedure.forth
			end
		end




feature -- Implementation


	postcondition: PRODUCTION_CLASS_NAVIGATOR



	while_loops: LINKED_SET [PRODUCTION_CLASS]



	generate_all_invariants (use_previous_values: BOOLEAN) is
			-- generate all candidate invariante for current postcondition
		   	-- use _only_ previous(IDs) iff use_previous_values = True
		local
			old_value: STRING
			new_invariant: PRODUCTION_CLASS_NAVIGATOR
			ids_handler, modified_handler: LINKED_SET [PRODUCTION_CLASS]
			a_while_loop: PRODUCTION_CLASS_NAVIGATOR
			modified, names_of_variables: LINKED_SET [STRING]
			all_mappings: MAPPINGS [PRODUCTION_CLASS, STRING]
			all_values, all_previous_modified: LINKED_SET [STRING]
		do
			new_invariant := postcondition.twin
			ids_handler := new_invariant.id_modification_handler
			create names_of_variables.make
			names_of_variables.compare_objects

			if name_based_substitution then
				-- collect all names of variables in the invariant
				from ids_handler.start
				until ids_handler.after
				loop
					names_of_variables.extend (ids_handler.item.value)
					ids_handler.forth
				end
			end

			-- the first generated invariant is simply the postcondition
			if not use_previous_values and variables_at_a_time = 0 then
				all_candidate_invariants.extend (postcondition.twin)
			end

			-- for every outer while loop
			from while_loops.start
			until while_loops.after
			loop
				create a_while_loop.make_with_modifies (while_loops.item, modifies_clauses)
				modified := a_while_loop.modified_variables
				if variables_at_a_time = 0 then
					-- for every modified variable
					from modified.start
					until modified.after
					loop
						if not name_based_substitution then
							-- for every variable in the invariant
							from ids_handler.start
							until ids_handler.after
							loop
								-- if the variable is not one of those modified in the loop
								if not modified.has (ids_handler.item.value) then
									-- replace unmodified variable with modified one
									old_value := ids_handler.item.value
									-- if use_previous_values
									if use_previous_values then
										ids_handler.item.set_value
										(previous(modified.item, a_while_loop))
									else
										ids_handler.item.set_value (modified.item)
									end
									-- add new invariant to set
									all_candidate_invariants.extend (new_invariant.twin)
									-- undo the modification to loop on
									ids_handler.item.set_value (old_value)
								end
								ids_handler.forth
							end
						else
							-- name_based_substitution
							-- for every variable ID in the invariant
							from names_of_variables.start
							until names_of_variables.after
							loop
								-- if the variable is not one of those modified in the loop
								if not modified.has (names_of_variables.item) then
									create modified_handler.make
									-- replace all occurrences of unmodified variable with modified one
									-- while keeping track of them in modified_handler
									from ids_handler.start
									until ids_handler.after
									loop
										if ids_handler.item.value.is_equal (names_of_variables.item) then
											modified_handler.extend (ids_handler.item)
											-- if use_previous_values
											if use_previous_values then
												ids_handler.item.set_value
												(previous(modified.item, a_while_loop))
											else
												ids_handler.item.set_value (modified.item)
											end
										end
										ids_handler.forth
									end
									-- add new invariant to set
									all_candidate_invariants.extend (new_invariant.twin)
									-- undo the modification to loop on
									from modified_handler.start
									until modified_handler.after
									loop
										modified_handler.item.set_value (names_of_variables.item)
										modified_handler.forth
									end
								end
								names_of_variables.forth
							end
						end
						modified.forth
					end
				else
					-- multiple_variables!

					-- save current values (for partial mappings)
					create all_values.make
					from ids_handler.start
					until ids_handler.after
					loop
						all_values.extend (ids_handler.item.value)
						ids_handler.forth
					end

					-- if try_previous_values add previous values to mappings
					create all_previous_modified.make
					if previous_values then
						from modified.start
						until modified.after
						loop
							all_previous_modified.extend (modified.item)
							all_previous_modified.extend (previous (modified.item, a_while_loop))
							modified.forth
						end
					else
						all_previous_modified.append (modified)
					end

					-- loop over all possible mappings
					create all_mappings.make_with_restricted_domain (ids_handler, all_previous_modified, variables_at_a_time)
					from all_mappings.start
					until all_mappings.after
					loop
						-- for every variable in the invariant
						from ids_handler.start
						until ids_handler.after
						loop
							if all_mappings.is_defined (ids_handler.item) then
								-- replace variable with what it maps to in the current mapping
								ids_handler.item.set_value (all_mappings.maps_to (ids_handler.item))
							end
							ids_handler.forth
						end
						-- add new invariant to set
						all_candidate_invariants.extend (new_invariant.twin)
						-- undo the modification to loop on
						from ids_handler.start; all_values.start
						until ids_handler.after
						loop
							ids_handler.item.set_value (all_values.item)
							ids_handler.forth
							all_values.forth
						end
						all_mappings.forth
					end
				end
				while_loops.forth
			end
		end




	previous (an_id: STRING; a_loop: PRODUCTION_CLASS): STRING is
			-- return a symbolic expression representing the value of variable
			-- with name `id' at the previous iteration of the block `a_loop'
		do
			-- dummy implementation
			Result := an_id.twin
			Result.append (" - 1")
		end

end
