-- Copyright (C) 2009  Carlo A. Furia
--                     caf@inf.ethz.ch  http://se.inf.ethz.ch/people/furia/
--                     
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.

class
   PRODUCTION_CLASS
	
inherit ANY
		redefine copy, is_equal end
		
	
	create
	make, make_with_value
	
	
	
feature -- Creation
	
	make (a_type: INTEGER)
		do
			type := a_type
			number_of_children := 0
			create children.make (number_of_children)
			reset (0)
		end
	
	
	make_with_value (a_type: INTEGER; a_value: STRING)
		do
			make (a_type)
			set_value (a_value)
		end
	
	
	
feature -- information
	
	
	type: INTEGER
	
	value: STRING
	
	number_of_children: INTEGER
	
	children: ARRAYED_LIST[PRODUCTION_CLASS]

	
	
	is_terminal: BOOLEAN
			-- does this class represent an input token?
		do
			Result := (number_of_children = 0)
		end
	
	
	add_child (a_child: PRODUCTION_CLASS)
		do
			children.extend (a_child)
			number_of_children := number_of_children + 1
		end
	
	
	set_value (a_value: STRING)
		-- sets the value of the current production
	   -- For terminal tokens, this is the semantic value that has 
	   -- been returned by the scanner
		do
			value := a_value
		end
	
	
	has_value: BOOLEAN
			-- does the production has a local value?
		do
			Result := value /= Void
		end


	
	
	
feature -- navigation
	
	item: PRODUCTION_CLASS
			-- Current node in the tree
	
	
	lookup_token: INTEGER
			-- token currently looked up during navigation
	
	count: INTEGER
			-- number of instances of token currently visited in subtree
	
	
	reset (token: INTEGER) is
			-- Set the current node to the root of the parse object
		   -- and set lookup_token to token
		do
			item := Current
			count := 0
			lookup_token := token
			from children.start
			until children.after
			loop children.item.reset (token); children.forth
			end
		end
	
	
	forth is
			-- move item to the next (in pre-order) occurrence
			-- of a node with type lookup_token
		   -- if such node does not exist item is set to Void
		local
			new_item: PRODUCTION_CLASS
			new_count: INTEGER
		do
			if type = lookup_token and count = 0 then
				new_item := Current
				new_count := 1
			else
				new_item := Void
				from children.start
				until children.after
				loop
					children.item.forth
					new_item := children.item.item
					new_count := new_count + children.item.count
					if new_item /= Void and count <= new_count then
						-- stop at first occurrence
						children.finish
					end
					children.forth
				end
			end
			item := new_item
			if item /= Void then
				count := count + 1
			end
		end
	
	
	back is
			-- move item to the previous (in pre-order) occurrence
			-- of a node with type lookup_token
			-- if such node does not exist item is set to Void
		do
			if count > 1 then
				i_th (lookup_token, count - 1)
			else
				item := Void
			end
		end
	
	
	after: BOOLEAN is
			-- are there no more occurrences of nodes with lookup_token?
		do
			Result := item = Void
		end
	
	
	start (token: INTEGER) is
			-- move item to the first (in pre-order) occurrence
			-- of a node with type token
		do
			reset (token)
			forth
		end
	
	
	finish (token: INTEGER) is
			-- move item to the last (in pre-order) occurrence
			-- of a node with type token
		do
			i_th_from_bottom (token, 1)
		end
	
	
	i_th (token: INTEGER; i: INTEGER) is
			-- move item to the i_th (in pre-order) occurrence
			-- of a node with type token
		require
			i > 0
		do
			from start (token)
			until count = i or after
			loop forth
			end
		end
	
	
	i_th_from_bottom (token: INTEGER; i: INTEGER) is
			-- move item to the i_th (from the bottom, in pre-order)
			-- occurrence of a node with type token
			-- i.e., if i = 1 return last element, if i = 2
			-- return second-to-last element, etc.
			-- item is Void if such occurrence does not exist
		require
			i > 0
		do
			-- determine total number of occurrences
			from start (token)
			until after
			loop forth
			end
			if count - i + 1 > 0 then
				i_th (token, count - i + 1)
			end
		end
	

	
	
	
feature -- copying and comparison
	
	copy (other: like Current) is
			-- Update current object using fields of other
		do
			if other.value = Void then
				make (other.type)
			else
				make_with_value (other.type, other.value.twin)
			end
			from other.children.start
			until other.children.after
			loop
				add_child (other.children.item.twin)
				other.children.forth
			end
		end
	
	
	is_equal (other: like Current):BOOLEAN is
			-- is other attached to an object considered
			-- equal to Current?
		do 
			Result := number_of_children = other.number_of_children
			   and ((value = Void and other.value = Void) or
					  (value /= Void and other.value /= Void and value.is_equal (other.value)))
			   and type = other.type
		end
	

end
