-- Copyright (C) 2009  Carlo A. Furia
--                     caf@inf.ethz.ch  http://se.inf.ethz.ch/people/furia/
--                     
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.


indexing

	description: "Navigator and pretty-printer of PRODUCTION_CLASS parse objects"

class PRODUCTION_CLASS_NAVIGATOR

inherit
	
	PRODUCTION_CLASS
		rename make as make_flat end
		
	BOOGIEPL_TOKEN_CODES
		undefine copy, is_equal end
	
	
	create make, make_with_modifies
	

feature {NONE}
	-- Creation
	
	make (parse_tree: PRODUCTION_CLASS) is
			-- set parse_tree as the object to navigate
		require
			parse_tree /= Void
		do
			number_of_children := parse_tree.number_of_children
			children := parse_tree.children
			value := parse_tree.value
			type := parse_tree.type
			reset (0)
			create_modifies_clauses
			reset (0)			
		ensure
			item /= Void
		end
	
	
	make_with_modifies (parse_tree: PRODUCTION_CLASS;
							  given_modifies_clauses: HASH_TABLE [LINKED_SET[STRING], STRING]) is
			-- set parse_tree as the object to navigate and 
			-- given_modifies_clauses as the modifies_clauses table
		require
			parse_tree /= Void
			given_modifies_clauses /= Void
		do
			make (parse_tree)
			modifies_clauses := given_modifies_clauses
		ensure
			item /= Void
			modifies_clauses /= Void
		end
	
	
	
feature
	-- Token-specific information collection
	
	outer_while_loops_handler: LINKED_SET [PRODUCTION_CLASS] is
			-- get a list of references to all PRODUCTION_CLASSes
			-- that encapsulate outer while loops
			-- all PRODUCTION_CLASSes in the returned list are of type
			-- Stmt and their children[1].type is WHILEKW
		do
			create Result.make
			Result.compare_references
			-- go to the next while block
			from start (Stmt)
			until after
			loop
				if item.children[1].type = WHILEKW then
					-- get a handle of while loop
					Result.extend (item)
					-- skip through all nested loops
					from item.start (Stmt)
					until item.after
					loop item.forth
					end
				end
				forth
			end
		end
		   
	
	
	id_modification_handler: LINKED_SET [PRODUCTION_CLASS] is
			-- get a list of references to all PRODUCTION_CLASSes
			-- that are found in the next Expr block and that can be
			-- modified with other variable ID
			-- all PRODUCTION_CLASSes in the returned list are of type
			-- IDID or NUM
		do
			-- go to the next Expr block
			start (Expr)
			-- go to the next E0 block
			Result := e_block_handler (item.children[1])
		end
	
	
	
	modifies (procedure: STRING): LINKED_SET [STRING] is
			-- get a list of ids present in the modifies
			-- clause of a procedure with name `procedure'.
			-- If no procedure with this name exists, return an empty list
		do
			if modifies_clauses = Void then
				-- first invocation: build table
				create_modifies_clauses
			end
			modifies_clauses.search (procedure)
			if modifies_clauses.found then
				Result := modifies_clauses.found_item.twin
			else
				Result := create {LINKED_SET [STRING]}.make
				Result.compare_objects
			end
		ensure
			Result /= Void
		end
	
	
	modified_variables: LINKED_SET [STRING] is
			-- get a list of ids that may be written to by the instructions
			-- in the next StmtList block of statements
		local
			statement, statement_list: PRODUCTION_CLASS
		do
			-- go to the next block of statements
			start (StmtList)
			if after then
				-- no list of statements available
				Result := Void
			else
				create Result.make
				Result.compare_objects
				-- get LStmtstar
				statement_list := item.children[1].twin
				from statement_list.start (LStmt)
				until statement_list.after
				loop
					if statement_list.item.number_of_children = 1 then
						statement := statement_list.item.children[1].twin
						-- add list of variables modified by statement to Result
						Result.append (modified_by_statement (statement))
					else
						-- labeled statement: discard labels
						-- this amounts to just skipping to the next iteration
					end
					statement_list.forth
				end
			end
		end
				


	
	
				
				
	
	
	
	
feature
	
	-- Pretty-printing
	
	print_leaves: STRING
	is
			-- Visit item in pre-order and return value fields of leaves
		do
			create result_string.make(0)
			indentation_level := 0
			last_semicolon := False
			last_closedcurly := False
			while_loop_condition := -1
			add_node_and_recur (item)
			Result := result_string
		end
	
	
	
feature {NONE}
	-- Implementation
	
	result_string: STRING
	
	indentation_level: INTEGER
	last_semicolon: BOOLEAN
	last_closedcurly: BOOLEAN
	while_loop_condition: INTEGER
	
	
	tab_until_indentation_level is
			-- print as many tab as the current value of indentation_level
		local
			level: INTEGER
		do
			from level := 1
			until level > indentation_level
			loop result_string.append("%T"); level := level + 1
			end
		end
			

	pretty_print_terminal (a_type: INTEGER; a_value: STRING)
		is
			-- Print a_value with pretty-printing based on the information in a_type
		   -- See boogiepl.y for reference
		do
			if a_type = SEMICOLONKW then
				result_string.append ( a_value + "%N")
				last_semicolon := True
				last_closedcurly := False
			elseif
				a_type = RETURNSKW
					or a_type = REQUIRESKW
					or a_type = ENSURESKW
					or a_type = MODIFIESKW then
					result_string.append ("%N%T" + a_value + " ")
					last_closedcurly := False
					last_semicolon := False
			elseif
				a_type = CURLYOPENKW then
				result_string.append ("%N")
				tab_until_indentation_level
				result_string.append (a_value + "%N")
				indentation_level := indentation_level + 1
				tab_until_indentation_level
				last_closedcurly := False
				last_semicolon := False
			elseif
				a_type = CURLYCLOSEDKW then
				if last_closedcurly then
					result_string.append ("%N")
				end
				indentation_level := indentation_level - 1
				tab_until_indentation_level
				result_string.append (a_value + "%N")
				tab_until_indentation_level
				last_closedcurly := True
				last_semicolon := False
			else
				if last_semicolon then
					tab_until_indentation_level
				end
				if a_type = WHILEKW then
					while_loop_condition := while_loop_condition + 1
				end
				if while_loop_condition > -1 and a_type = OPENPARENKW then
					while_loop_condition := while_loop_condition + 1
				end
				if while_loop_condition > -1 and a_type = CLOSEDPARENKW then
					while_loop_condition := while_loop_condition - 1
				end
				result_string.append (a_value + " ")
				last_closedcurly := False
				last_semicolon := False
				if while_loop_condition = 0 and a_type = CLOSEDPARENKW then
					result_string.append ("%N")
					tab_until_indentation_level
					while_loop_condition := -1
				end
			end
		end
	
	
	add_node_and_recur (node: PRODUCTION_CLASS) is
		do
		   if node /= Void  then
		      if node.is_terminal then
					if node.has_value then
						pretty_print_terminal (node.type, node.value)
					end
				else
					-- non-terminal node
					from node.children.start
					until node.children.after
					loop
						add_node_and_recur (node.children.item)
						node.children.forth
					end
				end
			end
		end
		
		
	modifies_clauses: HASH_TABLE [LINKED_SET[STRING], STRING]
	
	
	create_modifies_clauses is
			-- build the content of table modifies_clauses
		local
			specification: PRODUCTION_CLASS
			id_list: PRODUCTION_CLASS
			list_of_ids: LINKED_SET [STRING]
			current_procedure: STRING
		do
			create modifies_clauses.make (3)
			-- for every procedure declaration
			from start (ProcedureDecl)
			until after
			loop
				current_procedure := item.children[2].children[1].value.twin
				-- loop over all specifications:
				-- two possible productions of type ProcedureDecl
				if item.children[3].type = SEMICOLONKW then
					-- no implementation: pick last item
					specification := item.children[5].twin
				else
					-- implementation: pick 2nd-to-last item
					specification := item.children[4].twin
				end
				-- loop over all modifies clauses
				from specification.start (Spec)
				until specification.after
				loop
					-- look up modifies clauses
					if specification.item.children[1].type = MODIFIESKW then
						id_list := specification.item.children[2].twin
						-- collect list of ids in list_of_ids
						from
							id_list.start (IDID)
							create list_of_ids.make
							list_of_ids.compare_objects
						until id_list.after
						loop
							list_of_ids.extend (id_list.item.value)
							id_list.forth
						end
						-- see if current_procedure already has an associated 
						-- list modified ids
						if modifies_clauses /= Void then
							modifies_clauses.search (current_procedure)
							if modifies_clauses.found then
								-- other ids exist already for current_procedure
								modifies_clauses.found_item.append (list_of_ids)
								modifies_clauses.force (modifies_clauses.found_item.twin,
																current_procedure)
							else
								-- no other ids exist already for current_procedure
								modifies_clauses.put (list_of_ids, current_procedure)
							end
						end
					end
					specification.forth
				end
				forth
			end
		end
	
	
	modified_by_statement (statement: PRODUCTION_CLASS): LINKED_SET [STRING] is
			-- list of ids modified by statement
		require
			statement /= Void
			statement.type = Stmt
		local
			new_statement_navigator: PRODUCTION_CLASS_NAVIGATOR
			new_statement, block_A, block_B: PRODUCTION_CLASS
		do
			create Result.make
			Result.compare_objects
			if statement.children[1].type = HAVOCKW then
				-- Statement: havoc Idpluscomma
				-- Action: collect all IDID
				from statement.start (IDID)
				until statement.after
				loop
					Result.extend (statement.item.value)
					statement.forth
				end
			elseif
				statement.number_of_children > 1 and statement.children[2].type = ASSIGNEMENTKW then
				-- Statement: Lhspluscomma := ...
				-- Action: collect all IDID while ignoring MapSelectstar
				from statement.start (Lhs)
				until statement.after
				loop
					Result.extend (statement.item.children[1].children[1].value)
					statement.forth
				end
			elseif
				statement.children[1].type = CALLKW and statement.children[2].type = Id then
				-- Statement: procedure call without return values
				-- Action: get ids modified by called procedure	
				Result.append (modifies (statement.children[2].children[1].value))
			elseif
				statement.children[1].type = CALLKW and statement.children[2].type = CallLhs then
				-- Statement: procedure call with return values
				-- Action: get ids modified by called procedure...
				Result.append (modifies (statement.children[3].children[1].value))
				-- Action: ... and collect IDIDs in the Lhs of call instruction 				
				new_statement := statement.children[2]
				from new_statement.start (IDID)
				until new_statement.after
				loop
					Result.extend (new_statement.item.value)
					new_statement.forth
				end
			elseif
				statement.children[1].type = IfStmt then
				-- Statement: if ...
				-- Action: recur on then and else part
				new_statement := statement.children[1]
				create new_statement_navigator.make_with_modifies (new_statement.children[5].children[2].twin,
																					modifies_clauses)
				Result.append (new_statement_navigator.modified_variables)
				if new_statement.number_of_children = 6 then
					-- else part is present
					new_statement := new_statement.children[6]
					if new_statement.children[2].type = BlockStmt then
						create new_statement_navigator.make_with_modifies (new_statement.children[2].children[2],
																							modifies_clauses)
						Result.append (new_statement_navigator.modified_variables)
					else
						check new_statement.children[2].type = IfStmt end
						-- building suitable nested structure, bottom up					
						-- build Stmt node
						create block_A.make (Stmt)
						-- add IfStmt node below it
						block_A.add_child (new_statement.children[2].twin)
						-- build LStmt node
						create block_B.make (LStmt)
						-- add Stmt node below it
						block_B.add_child (block_A)
						-- build LStmtstar node
						create block_A.make (LStmtstar)
						-- add LStmtstar empty node below it
						block_A.add_child (create {PRODUCTION_CLASS}.make (LStmtstar))
						-- add Lstmt node below it
						block_A.add_child (block_B)
						-- build StmtList node
						create block_B.make (StmtList)
						-- add LStmtstar node below it
						block_B.add_child (block_A)
						-- now block_B can be navigated
						create new_statement_navigator.make_with_modifies (block_B, modifies_clauses)
						Result.append (new_statement_navigator.modified_variables)
					end
				end
			elseif
				statement.children[1].type = WHILEKW then
				-- Statement: while ...
				-- Action: recur on then loop body
				new_statement := statement.children[6]
				create new_statement_navigator.make_with_modifies (new_statement.children[2].twin,
																					modifies_clauses)
				Result.append (new_statement_navigator.modified_variables)
			end
		end
	
	
	
	e_block_handler (e_block: PRODUCTION_CLASS): LINKED_SET [PRODUCTION_CLASS] is
			-- get a list of references to all PRODUCTION_CLASSes			
			-- that are found in the e_block and that can be
			-- modified with other variable ID
			-- all PRODUCTION_CLASSes in the returned list are of type
			-- IDID or NUM
		   -- e_block must be some of type Expr or Ei for 0 <= i <= 9
		require
			e_block.type = Expr
				or	e_block.type = E0
				or	e_block.type = E1
				or	e_block.type = E2			
				or	e_block.type = E3
				or	e_block.type = E4
				or	e_block.type = E5
				or	e_block.type = E6
				or	e_block.type = E7
				or	e_block.type = E8
				or	e_block.type = E9
		local
			block, new_block, other_block: PRODUCTION_CLASS
			new_ids: LINKED_SET [PRODUCTION_CLASS]
			quantified_ids: LINKED_SET [STRING]
		do
			if e_block.type = Expr then
				-- Expr
				-- Action: recur on E0
				Result := e_block_handler (e_block.children[1])
			elseif
				e_block.type = E0
					or e_block.type = E1
					or e_block.type = E3
					or e_block.type = E4
					or e_block.type = E5
					or e_block.type = E6 then
					-- E0, E1, E3, E4, E5, or E6
					-- Action: recur on children[1] and children[3] if it exists
					Result := e_block_handler (e_block.children[1])
				if e_block.number_of_children = 3 then
					Result.append (e_block_handler (e_block.children[3]))
				end
			elseif
				e_block.type = E2 then
				-- E2
				-- Action: recur on E3 and other E3 block if it exists
				Result := e_block_handler (e_block.children[1])
				if e_block.number_of_children = 2 then
					from new_block := e_block.children[2]
					until new_block.number_of_children = 1
					loop
						Result.append (e_block_handler (new_block.children[2].children[2]))
						new_block := new_block.children[1]
					end
					Result.append (e_block_handler (new_block.children[1].children[2]))
				end
			elseif
				e_block.type = E7 then
				-- E7
				-- Action: recur on E8
				Result := e_block_handler (e_block.children[2])
			elseif
				e_block.type = E8 then
				-- E8
				-- Action: recur on E9 if no MapOp present, otherwise recur
				-- on argument of map operators
				if e_block.children[2].number_of_children = 0 then
					-- no MapOp present
					Result := e_block_handler (e_block.children[1])
				else
					-- MapOp present: recur on arguments
					create Result.make
					Result.compare_references
					from new_block := e_block.children[2]
					until new_block.number_of_children = 0
					loop
						if new_block.children[2].children[2].type = Exprpluscomma then
							from other_block := new_block.children[2].children[2]
							until other_block.number_of_children = 1
							loop
								-- add Expr
								Result.append (e_block_handler (other_block.children[3]))
								-- move to next Exprpluscomma
								other_block := other_block.children[1]
							end
							-- add last Expr
							Result.append (e_block_handler (other_block.children[1]))
							if new_block.children[2].children[3].type = MapUpdate then
								-- handle MapUpdate as well
								other_block := new_block.children[2].children[3]
								Result.append (e_block_handler (other_block.children[2]))
							end
						end
						new_block := new_block.children[1]
					end
				end
			elseif
				e_block.type = E9 then
				-- E9
				-- Action: handle various cases
				create Result.make
				Result.compare_references				
				if e_block.children[1].type = FALSEKW
					or e_block.children[1].type = TRUEKW then
					Result.extend (e_block.children[1])
				elseif e_block.children[1].type = Number then
					Result.extend (e_block.children[1].children[1])
				elseif
					e_block.children[1].type = Id then
					if e_block.number_of_children = 1 then
						-- just IDID: simply add it
						Result.extend (e_block.children[1].children[1])
					else
						-- function application present
						-- recur on arguments (if any)
						new_block := e_block.children[2].children[2]
						if new_block.number_of_children = 1 then
							from new_block := new_block.children[1]
							until new_block.number_of_children = 1
							loop
								Result.append (e_block_handler (new_block.children[3]))
								new_block := new_block.children[1]
							end
							Result.append (e_block_handler (new_block.children[1]))
						end
					end
				elseif
					e_block.children[1].type = OLDKW then
					-- old ( Expr ), recur on argument
					Result := e_block_handler (e_block.children[3])
				elseif
					e_block.number_of_children > 5 then
					-- quantified expression
					if e_block.children[3].type = TypeArgs then
						new_block := e_block.children[6]
						other_block := e_block.children[4]
					else
						new_block := e_block.children[5]
						other_block := e_block.children[3]
					end
					-- collect new candidate IDs by recurring
					new_ids := e_block_handler (new_block)
					-- collect IDs that represent quantified variables
					create quantified_ids.make
					quantified_ids.compare_objects
					from block := other_block.twin; block.start (IDID)
					until block.after
					loop
						quantified_ids.extend (block.item.value)
						block.forth
					end
					-- add only those IDs that are not quantified variables
					from new_ids.start
					until new_ids.after
					loop
						if not quantified_ids.has(new_ids.item.value) then
							Result.extend (new_ids.item)
						end
						new_ids.forth
					end
				else
					-- ( Expr )
					-- recur on Expr
					Result := e_block_handler (e_block.children[2])
				end
			end
		end
		     
		     
	
	
end
